package jp.gr.java_conf.matsubara.expresser;

import jp.gr.java_conf.matsubara.expresser.exception.ExpException;
import jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression;
import jp.gr.java_conf.matsubara.expresser.type.IExpObject;
import jp.gr.java_conf.matsubara.expresser.type._Boolean;

/**
 * RZqiop1 ? op2 : op3j<br/>
 * <br/>
 * 쐬: 2005/06/04
 * 
 * @author m.matsubara
 */
class SelectParsedExpression implements IParsedExpression {
    private IParsedExpression operand1;
    private IParsedExpression operand2;
    private IParsedExpression operand3;
    
    public SelectParsedExpression(IParsedExpression operand1, IParsedExpression operand2, IParsedExpression operand3) {
        this.operand1 = operand1;
        this.operand2 = operand2;
        this.operand3 = operand3;
    }
    
    /* ( Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#evaluate(jp.gr.java_conf.matsubara.expresser.RuntimeData)
     */
    public IExpObject evaluate(RuntimeData runtime) throws ExpException {
        IExpObject value = operand1.evaluate(runtime);
        _Boolean bool;
        if (value instanceof _Boolean)
            bool = (_Boolean)value;
        else
            bool = runtime.getTypeConverter().toBoolean(runtime, value);
        if (bool.booleanValue()) {
            return operand2.evaluate(runtime);
        } else {
            return operand3.evaluate(runtime);
        }
    }
    
    /* (non-Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#evalRef(jp.gr.java_conf.matsubara.expresser.RuntimeData)
     */
    public IVariable evalRef(RuntimeData runtime) throws ExpException {
        throw new ExpException(Expresser.getResString("NotSubstituteToConstant"), -1);
    }
    
    /* (non-Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#evalTargetObject(jp.gr.java_conf.matsubara.expresser.RuntimeData)
     */
    public IExpObject evalTargetObject(RuntimeData runtime) throws ExpException {
        return Expresser.UNDEFINED;
    }
    
    /* ( Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#isConst()
     */
    public boolean isConst() {
        //  {operand1̒lɉoperand2, operand3͕Еł悢̂ARuntimeDataIuWFNgȂ̂Ń`FbNłȂ
        //  operand1, operand2, operand3 SĂ萔̏ȂĒNȂƎvcB
        return operand1.isConst() && operand2.isConst() && operand3.isConst(); 
    }

    /* ( Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#isUsingVariable(java.lang.Integer)
     */
    public int isUsingVariable(Integer id) throws ExpException {
        return operand1.isUsingVariable(id) + operand2.isUsingVariable(id) + operand3.isUsingVariable(id);
    }
}



/**
 * RZqiop1 ? op2 : op3j̃p[T[
 * 
 * 쐬: 2005/05/29
 * 
 * @author m.matsubara
 */
public class SelectOperatorParser implements IOperatorParser {

    /* ( Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.IOperatorParser#parseOperand2(Expresser expresser, char[], int, int)
     */
    public IParsedExpression parse(Expresser expresser, char[] statement, int begin,
            int end, int opeIdx, String opeWord) throws ExpException {
        IParsedExpression operand1 = expresser.parse(statement, begin, opeIdx);
        
        Range[] selectValues = Expresser.split(':', statement, opeIdx + opeWord.length(), end);
        if (selectValues.length != 2)
            throw new ExpException(Expresser.getResString("SyntaxError"), opeIdx);

        IParsedExpression operand2 = expresser.parse(statement, selectValues[0]);
        IParsedExpression operand3 = expresser.parse(statement, selectValues[1]);
        
        return new SelectParsedExpression(operand1, operand2, operand3);
    }

}
