package jp.gr.java_conf.matsubara.expresser.parsedexpression;

import jp.gr.java_conf.matsubara.expresser.Expresser;
import jp.gr.java_conf.matsubara.expresser.IVariable;
import jp.gr.java_conf.matsubara.expresser.RuntimeData;
import jp.gr.java_conf.matsubara.expresser.exception.ExpException;
import jp.gr.java_conf.matsubara.expresser.type.ExpObjectMember;
import jp.gr.java_conf.matsubara.expresser.type.IExpObject;

/**
 * o[ANZX\܂B<br/>
 * o[ANZXZqOperatorSetł͂ȂpIParsedExpressionIuWFNgɂď܂B<br/> 
 * <br/>
 * 쐬: 2005/07/29
 * 
 * @author m.matsubara
 */
public class Op_MemberAccessExpression implements IParsedExpression {
    private IParsedExpression objectExpression;
    private Integer memberID;
    private int opeIdx;
    
    public Op_MemberAccessExpression(IParsedExpression arrayExpression, Integer memberID, int opeIdx) {
        this.objectExpression = arrayExpression;
        this.memberID = memberID;
        this.opeIdx = opeIdx;
    }
    
    /*
     *  (non-Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#isUsingVariable(java.lang.Integer)
     */
    public int isUsingVariable(Integer id) throws ExpException {
        return objectExpression.isUsingVariable(id);
    }

    /*
     *  (non-Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#evaluate(jp.gr.java_conf.matsubara.expresser.RuntimeData)
     */
    public IExpObject evaluate(RuntimeData runtime) throws ExpException {
        try {
            IExpObject object = objectExpression.evaluate(runtime);
            if (object == null)
                throw new ExpException(Expresser.getResString("NullObject"), -1);
            return object.getProperty(memberID);
        } catch (ExpException ee) {
            if (ee.getPosition() < 0)
                ee.setPosition(opeIdx);
            throw ee;
        }
    }

    /* (non-Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#evalRef(jp.gr.java_conf.matsubara.expresser.RuntimeData)
     */
    public IVariable evalRef(RuntimeData runtime) throws ExpException {
        try {
            IExpObject object = objectExpression.evaluate(runtime);
            if (object == null)
                throw new ExpException(Expresser.getResString("NullObject"), -1);
            return new ExpObjectMember(object, memberID);
        } catch (ExpException ee) {
            if (ee.getPosition() < 0)
                ee.setPosition(opeIdx);
            throw ee;
        }
    }
    
    /* (non-Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#evalTargetObject(jp.gr.java_conf.matsubara.expresser.RuntimeData)
     */
    public IExpObject evalTargetObject(RuntimeData runtime) throws ExpException {
        IExpObject object = objectExpression.evaluate(runtime);
        return object;
    }
    
    /*
     *  (non-Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.parsedexpression.IParsedExpression#isConst()
     */
    public boolean isConst() {
        return false;
    }
}
