package jp.gr.java_conf.matsubara.interpreter.ecmascript.parser;

import jp.gr.java_conf.matsubara.expresser.Expresser;
import jp.gr.java_conf.matsubara.expresser.Range;
import jp.gr.java_conf.matsubara.expresser.StringToId;
import jp.gr.java_conf.matsubara.expresser.exception.ExpException;
import jp.gr.java_conf.matsubara.interpreter.ScriptEngine;
import jp.gr.java_conf.matsubara.interpreter.statement.BlockStatement;
import jp.gr.java_conf.matsubara.interpreter.statement.FunctionStatement;
import jp.gr.java_conf.matsubara.interpreter.statement.IStatement;

/**
 * Function Xe[ggi֐̒`jłB<br/>
 * 
 * 쐬: 2005/02/03
 * 
 * @author m.matsubara
 *
 */
public class FunctionStatementParser implements IStatementParser {
    public String targetStatement() {
        return "function";
    }
    
    /* (Javadoc Ȃ)
     * @see jp.gr.java_conf.matsubara.interpreter.ecmascript.parser.IStatementParser#parseStatement(jp.gr.java_conf.matsubara.interpreter.ScriptEngine, jp.gr.java_conf.matsubara.interpreter.statement.BlockStatement, java.lang.String, char[], int, int, java.lang.Integer)
     */
    public int parseStatement(ScriptEngine se, BlockStatement blockStatement, String statementWord, char[] script, int begin, int end, Integer labelID) throws ExpException {
        int idx = begin;
        int statementLength = statementWord.length();

        idx = Expresser.skipWhitespace(script, idx + statementLength, end);
        String functionName = "";
        
        //  ֐擾
        if (script[idx] != '(') {
            functionName = se.getExpresser().nextToken(script, idx, end);
            idx += functionName.length();
            idx = Expresser.skipWhitespace(script, idx, end);
        }
        
        //  ֐͂
        if (script[idx] != '(')
            throw new ExpException(ScriptEngine.getResString("NecessaryParentheses"), idx);
        int argBegin = idx;
        idx = Expresser.skipParenthesis(script, idx, end);
        int argEnd = idx; 
        Range[] range = Expresser.split(',', script, argBegin + 1, argEnd - 1);
        int max = range.length; 

        //  B̍ڂȂ폜
        if (max == 1 && (range[0].getBegin() == range[0].getEnd()))
            max = 0;
        
        Integer[] argIds = new Integer[max];
        for (int argIdx = 0; argIdx < max; argIdx++) {
            int beginIdx = range[argIdx].getBegin();
            int endIdx   = range[argIdx].getEnd();
            String argName = new String(script, beginIdx, endIdx - beginIdx);
            if (se.getExpresser().isIdentifier(argName) == false)
                throw new ExpException(ScriptEngine.getResString("IllegalIdentifier", argName), beginIdx); 
            argIds[argIdx] = StringToId.toId(argName);
        }
        idx = Expresser.skipWhitespace(script, idx, end);
        
        //  ֐͂
        if (script[idx] != '{')
            throw new ExpException(ScriptEngine.getResString("FunctionBodyNotFound"), idx);
        int funcBodyBegin = idx + 1;
        idx = Expresser.skipParenthesis(script, idx, end);
        int funcBodyEnd = idx - 1;
        IStatement functionBody = se.parseStatement(script, funcBodyBegin, funcBodyEnd);
        FunctionStatement functionStatement;
        if ("".equals(functionName) == false)
            functionStatement = new FunctionStatement(functionName, StringToId.toId(functionName), argIds);
        else
            functionStatement = new FunctionStatement(functionName, null, argIds);
            
        //  ֐{̂̒ǉ
        int functionBodySize = ((BlockStatement)functionBody).size();
        if (functionBodySize == 1)
            functionStatement.setStatement(((BlockStatement)functionBody).getStatement(0));
        else
            functionStatement.setStatement(functionBody);    
        
        blockStatement.add(functionStatement);
        
        return idx;
    }
}
